#include <ExtendedLog/EntryType.hpp>
#include <ExtendedLog/LogWriter.hpp>
#include <QByteArray>
#include <QCloseEvent>
#include <QDateTime>
#include <QDir>
#include <QFile>
#include <QFileInfoList>
#include <QMessageBox>
#include <QStringList>
#include <tuple>

#include "exif.hpp"
#include "rename.hpp"

Rename::Rename(QString path, int minutes, int hours, ExtendedLog::LogWriter *logWriter, QObject *parent) :
	QObject{parent},
	dir(path),
	minutes(minutes),
	hours(hours),
	logWriter(logWriter),
	countedNames()
{

}

Rename::~Rename()
{
	delete logWriter;
}

QString Rename::createCountName(QString file, int counter)
{
	return file.left(file.length() - 4) + "_" + QString::number(counter) + ".jpg";
}

void Rename::rename()
{
	if (logWriter == nullptr)
	{
		QString errorMessage = tr("No log initialized.");

		/*
		 * We have to signal that somethig goes wrong. Of course when
		 * `logWriter` is a nullpointer we must not use it. Luckily we emit an
		 * `error` signal that is caught by our progress windows.
		 */
		emit error(errorMessage);
		return;
	}

	if (dir == "")
	{
		QString errorMessage = tr("No path selected!");
		logWriter->writeEntry(extLog::EntryType::ERROR, errorMessage);
		emit error(errorMessage);
		return;
	}

	logWriter->writeEntry(extLog::EntryType::INFORMATION, tr("%1 selected as path.").arg(dir));

	QDir working_dir(dir);
	working_dir.setNameFilters(filters);
	working_dir.setFilter(QDir::Files);

	auto files = working_dir.entryInfoList();

	numberOfFiles = files.count();
	int currentFiles = 0;
	int errorCount = 0;

	for (auto &&file : files)
	{
		QString fullFileName = file.canonicalFilePath(); // full filename (incl. path)
		QString fileName = file.fileName(); // filename without path

		currentFiles++;

		// Logging
		logWriter->writeEntry(extLog::EntryType::INFORMATION, tr("%1 selected.").arg(fileName));

		// check for JPEG-file not necessary because of filtering !?!

		QString message = "";
		QString newFullName = "";
		QString newName = "";

		QFile photoFile(fullFileName);
		if (photoFile.open(QIODevice::ReadWrite))
		{
			QByteArray data = photoFile.readAll();
			easyexif::EXIFInfo exifInfo;
			if (int code = exifInfo.parseFrom((unsigned char*)data.data(), data.size()))
			{
                logWriter->writeEntry(
                                extLog::EntryType::ERROR,
								tr("Error parsing EXIF-Data from %1: Code %2").arg(fileName).arg(code)
                           );

                message = tr("Error parsing EXIF-Data: Code %1").arg(code);
			}
			else
			{
                std::string exifDateTime = exifInfo.DateTimeOriginal != "" ? exifInfo.DateTimeOriginal : exifInfo.DateTime;

                // only rename iff we have date and time information in the file
                if (exifDateTime != "")
                {
                    QString dateTimeString = correctDateTimeFormat((QString::fromStdString(exifDateTime)));
                    QDateTime dateTime = QDateTime::fromString(dateTimeString, Qt::ISODate);

					newName = newFileName(dateTime, hours, minutes);

					QString pathSeperator = dir.endsWith('/')? "" : "/";
					newFullName = dir + pathSeperator + newName;
                    logWriter->writeEntry(
                                extLog::EntryType::INFORMATION,
                                tr("Date %1 found.").arg(dateTime.toString("dd.MM.yyyy hh:mm:ss"))
                            );

                    // renaming
                    int count = 1;
					QString fullCountName = newFullName;
					QString countName = newName;
					if (newFullName == fullFileName)
                    {
                        message = tr("File %1 has already the correct name.").arg(fileName);
                        logWriter->writeEntry(extLog::EntryType::INFORMATION, message);
                    }
                    else
                    {
						while (QFile::exists(fullCountName))
                        {
							QString fullCountNameTmp = createCountName(newFullName, ++count);
							QString countNameTmp = createCountName(newName, count);
							message = tr("File %1 already exists. Trying %2 as new filename.").arg(countName, countNameTmp);
                            logWriter->writeEntry(extLog::EntryType::WARNING, message);
							fullCountName = fullCountNameTmp;
							countName = countNameTmp;
                        }
						if (photoFile.rename(fullCountName))
                        {
							message = fileName + " -> " + fullCountName;
                            logWriter->writeEntry(
                                        extLog::EntryType::INFORMATION,
										tr("File %1 renamed to %2.").arg(fileName, countName)
                                    );

							/*
							 * check if the name of the new file is longer than 15 characters.
							 * if so, push this name to the list of counted filenames
							 */
							QString newFileName = fullCountName.split("/").last();
							newFileName.chop(4);
							int nameLenght = newFileName.length();
							if (nameLenght > 15) { countedNames.insert(newFileName.left(15));}
                        }
                        else
                        {
                            message = tr("Rename %1 failed.").arg(fileName);
                            logWriter->writeEntry(extLog::EntryType::INFORMATION, message);
                            errorCount++;
                        }
                    }

                }
                // no information -> we do nothing
                else
                {
                    message = tr("No EXIF-Data found.");
                }
            }
		}

		QStringList userState;
		userState << QString::number(currentFiles) << QString::number(numberOfFiles + countedNames.size()) << message;
		emit progressChanged(getPercent(currentFiles), userState);
	}
	logWriter->writeEntry(extLog::EntryType::INFORMATION, tr("First step completed with %n error(s).", "", errorCount));
	correctBaseNames();
}

QString Rename::newFileName(QDateTime datetime, int hr_plus, int min_plus)
{
	QString new_filename = "";
	qlonglong adjusting = (hr_plus * 3600) + (min_plus * 60);
	QDateTime datetime_adjusted = datetime.addSecs(adjusting);
	new_filename += QString::number(datetime_adjusted.date().year()); // for the year we do not need to add padding zeros
	new_filename += addZero(datetime_adjusted.date().month());
	new_filename += addZero(datetime_adjusted.date().day());
	new_filename += "_";
	new_filename += addZero(datetime_adjusted.time().hour());
	new_filename += addZero(datetime_adjusted.time().minute());
	new_filename += addZero(datetime_adjusted.time().second());
	new_filename += ".jpg";
	return new_filename;
}

QString Rename::addZero(int value)
{
	QString stringValue = QString::number(value);
	return (value >= 0 && value < 10) ? QString::number(0) + stringValue : stringValue;
}

int Rename::getPercent(double curr)
{
	/*
	 * percentage is number of currently vistied files in relation to the toatal number of files
	 * plus the files that must be visited a second time due to the counter-sorting-problem
	 */
	double percent = curr / (double)(numberOfFiles  + countedNames.size()) * 100;
	return (int)percent;
}

QString Rename::correctDateTimeFormat(QString fromExifLibrary)
{
	QString result = fromExifLibrary;
    if (result != "")
    {
        result.replace(4, 1, '-');
        result.replace(7, 1, '-');
        result.replace(10, 1, 'T');
    }
	return result;
}

void Rename::correctBaseNames()
{
	QString directory = QString(dir);
#if defined(Q_OS_WIN)
		if (!dir.endsWith("\\")) { directory.append("\\"); }
#elif defined(Q_OS_UNIX)
		if (!dir.endsWith("/")) { directory.append("/"); }
#endif
	logWriter->writeEntry(extLog::EntryType::INFORMATION, tr("Second step: correct sorting for counted files."));
	int length = countedNames.size();
	size_t counter = 0;
	for (auto &&name : countedNames)
	{
		counter++;
		QString notCountedName = QString(name + ".jpg");
		QString countedName = QString(name + "_1" + ".jpg");
		QString fullNotCountedName = QString(directory + notCountedName);
		QString fullCountedName = QString(directory + countedName);

		QStringList state;
		state
				<< QString::number(numberOfFiles + counter)
				<< QString::number(numberOfFiles + length)
				<< notCountedName + " -> " + countedName;
		emit progressChanged(getPercent(counter), state);

		std::tuple<QString, extLog::EntryType> logValues =
				 QFile::rename(fullNotCountedName, fullCountedName)
				? std::make_tuple(tr("File %1 renamed to %2.").arg(notCountedName).arg(countedName), extLog::EntryType::INFORMATION)
				: std::make_tuple(tr("Renaming %1 failed.").arg(notCountedName), extLog::EntryType::WARNING);

		logWriter->writeEntry(std::get<extLog::EntryType>(logValues), std::get<QString>(logValues));
	}

	logWriter->writeEntry(extLog::EntryType::INFORMATION, tr("Second step completed."));
	logWriter->close();
	emit finished();
}

